package cs4099;

import javax.swing.*;
import java.awt.*;

/**
 * This class is an extension of JPanel that displays cards in a pile arrangement.
 * The view can either be plain or similar to realistic cards.
 * 
 * @author David Newton
 * @see Game
 */
public class Desktop extends JPanel {
	
	protected State state;
	private final int VSEPARATION = 16;
	private final int CARDHEIGHT = 48;
	
	/**
	 * Constructs a new Desktop.
	 * 
	 * @param rootstate The initial state to draw.
	 * @param cardview True if the cards should be displayed, False if only suits and values should be visible.
	 */
	public Desktop(State rootstate, boolean cardview) {
		this.setBackground(Color.WHITE);
		this.setLayout(new AbsoluteLayout());
		this.state = rootstate;
		if (cardview) {
			drawPilesWithCards(rootstate);
		}
		else {
			drawPiles(rootstate);
		}
	}
	
	/**
	 * Draws the piles of the given state in plain form.
	 * @param state The state to draw.
	 */
	public void drawPiles(State state) {
		this.removeAll();
		this.setBackground(Color.getHSBColor(0f, 0f, 1f));
		int i = 0;
		//While there are more piles to draw...
		while (i < state.piles.size()) {
			
			Pile pile = ((Pile)(state.piles).get(i));
			
			//Get the X and Y coordinates to start from.
			int x = pile.displayx;
			int y = pile.displayy;
			
			//Draw a line representing the pile's start.
			if (pile.isfoundation) {
				this.add(new JLabel("*" + i + "*"), new AbsoluteConstraints(x, y, 32, VSEPARATION));
			}
			else if (pile.isstock) {
				this.add(new JLabel("S" + i), new AbsoluteConstraints(x, y, 32, VSEPARATION));
			}
			else {
				this.add(new JLabel("-" + i + "-"), new AbsoluteConstraints(x, y, 32, VSEPARATION));
			}
			
			//Now draw any cards on the pile.
			//Modify the VSEPARATION if the pile is to be displayed backwards.
			int ydistance;
			if (pile.backwards) {
				ydistance = 0 - VSEPARATION;
			}
			else {
				ydistance = VSEPARATION;
			}
			
			//Easy if it's a straight stack. Use the top card.
			if (pile.stackpointer != 0) {
				if (pile.straightstack) {
					Card card = new Card(pile.getTopCard().suit, pile.getTopCard().value);
					JLabel label = card.getCardLabel();
					this.add(label, new AbsoluteConstraints(x, y+ydistance, 32, 16));
				}
				
				else {
					//Iterate through all cards in the pile and add their representations.
					int j = Math.max(0, pile.stackpointer - 25);
					int offset = 0;
					if (j > 0) {
						this.add(new JLabel(". . ."), new AbsoluteConstraints(x, y+((offset+1)*ydistance), 32, VSEPARATION));
						offset++;
					}
					while (j < pile.stackpointer) {
						Card card = new Card(pile.contents[j].suit, pile.contents[j].value);
						JLabel label = card.getCardLabel();
						this.add(label, new AbsoluteConstraints(x, y+((offset)*ydistance), 32, CARDHEIGHT));
						j++;
						offset++;
					}
				}
			}
			i++;
		}
	}
	
	/**
	 * Draws a state using images that resemble cards.
	 * @param state The state to draw.
	 */
	public void drawPilesWithCards(State state) {
		this.removeAll();
		this.setBackground(Color.getHSBColor(0.25f, 0.6f, 0.5f));
		int i = 0;
		//While there are more piles left to draw...
		while (i < state.piles.size()) {
			Pile pile = ((Pile)(state.piles).get(i));
			
			//Get the X and Y coordinates to start from.
			int x = pile.displayx;
			int y = pile.displayy;
			
			//Draw a line representing the pile's start.
			if (pile.isfoundation) {
				this.add(new JLabel("*" + i + "*", JLabel.CENTER), new AbsoluteConstraints(x, y, 32, VSEPARATION));
			}
			else if (pile.isstock) {
				this.add(new JLabel("S" + i, JLabel.CENTER), new AbsoluteConstraints(x, y, 32, VSEPARATION));
			}
			else {
				this.add(new JLabel("-" + i + "-", JLabel.CENTER), new AbsoluteConstraints(x, y, 32, VSEPARATION));
			}
			
//			Now draw any cards on the pile.
			//Modify the VSEPARATION if the pile is to be displayed backwards.
			int ydistance;
			if (pile.backwards) {
				ydistance = 0 - VSEPARATION;
			}
			else {
				ydistance = VSEPARATION;
			}
			
			//Easy if it's a straight stack. Use the top card.
			if (pile.stackpointer != 0) {
				if (pile.straightstack) {
					Card card = new Card(pile.getTopCard().suit, pile.getTopCard().value);
					JLabel label = card.getCardImage();
					this.add(label, new AbsoluteConstraints(x, y+ydistance, 32, CARDHEIGHT));
				}
				else {
					//Iterate through all cards in the pile and add their representations.
					//For the card view, this must be done from the top down.
					//Start the marker from the furthest point out.
					if (pile.stackpointer > 25) {
						ydistance = ydistance / 2;
					}
					int j = pile.stackpointer-1;
					while (j >= 0) {
						JLabel label = pile.contents[j].getCardImage();
						this.add(label, new AbsoluteConstraints(x, y+8+((j+1)*ydistance), 32, CARDHEIGHT));
						j--;
					}
				}
			}
			i++;
		}
	}

}
