package cs4099;

//File handling
import java.io.File;
import java.io.IOException;

//Parsing
import javax.xml.parsers.DocumentBuilder; 
import javax.xml.parsers.DocumentBuilderFactory;  
import javax.xml.parsers.ParserConfigurationException;

//Document Object Model
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;

//Exceptions
import org.xml.sax.SAXException; 

import java.util.*;
import javax.swing.*;

/**
 * This class interprets an XML file for use in the main program. If the XML file does
 * not conform to the expected format, an error message will be displayed.<br/><br/>
 * 
 * The purpose of the code is to construct a root state from which a game is to be played.
 * 
 * @author David Newton
 * @see State
 */
public class Interpreter {
	
	/**
	 * A large method to go through the entire process of reading the properties of the XML file (game,
	 * piles, fixed cards and tactics) and storing them in a State ready for use by the program.
	 * 
	 * @param filename The file to interpret.
	 * @return A state matching the specification given in the file.
	 */
	public State setupGame(String filename) {
   	
//		Start with a blank state, then read the piles from the file and add them to it one by one.
		State state = new State(null);
    	DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();

    	try {
        	factory.setNamespaceAware(true);
        
    		DocumentBuilder builder = factory.newDocumentBuilder();
    		Document document = builder.parse ( new File("./xmls/!blank.xml"));
    		try {
    			document = builder.parse( new File(filename) );
    		}
    		catch (Exception e) {
    			JOptionPane.showMessageDialog(null, "The specified file could not be found.", "File Not Found", JOptionPane.ERROR_MESSAGE);
    		}
    		
    		state.gamename = document.getElementsByTagName("game").item(0).getAttributes().getNamedItem("name").getNodeValue();
    		
    		int deckstouse = 1;
    		try {
    			deckstouse = Integer.parseInt(document.getElementsByTagName("game").item(0).getAttributes().getNamedItem("decks").getNodeValue());
    		}
    		catch (Exception e) {
    		}
    		NodeList pilestoadd = document.getElementsByTagName("pile");
    		
    		LinkedList pilestoburykings = new LinkedList();
    		
    		//For each pile found in the XML file...
    		for (int i = 0; i < pilestoadd.getLength(); i++) {
    			
    			//Put the attributes of the node into a NamedNodeMap.
    			NamedNodeMap piledata = (pilestoadd.item(i)).getAttributes();
    			
    			//Read these attributes into variables
    			int id = intRead(piledata, "id", 0);
    			int capacity = intRead(piledata, "capacity", 52);
    			int order = intRead(piledata, "order", 0);
    			int uniquesuit = intRead(piledata, "uniquesuit", -1);
    			int stockcapacity = intRead(piledata, "stockcapacity", 1);
    			int startvalue = intRead(piledata, "startvalue", 0);
    			int cardstodeal = intRead(piledata, "cardstodeal", 0);
    			int stocknumber = intRead(piledata, "stocknumber", -1);
    			boolean alternatecolours = booleanRead(piledata, "alternatecolours", false);
    			boolean samesuit = booleanRead(piledata, "samesuit", false);
    			boolean removable = booleanRead(piledata, "removable", false);
    			boolean addable = booleanRead(piledata, "addable", false);
    			boolean isfoundation = booleanRead(piledata, "isfoundation", false);
    			boolean isstock = booleanRead(piledata, "isstock", false);
    			boolean oneaway = booleanRead(piledata, "oneaway", false);
    			boolean oneawaynowrap = booleanRead(piledata, "oneawaynowrap", false);
    			boolean backwards = booleanRead(piledata, "backwards", false);
    			boolean fourteens = booleanRead(piledata, "fourteens", false);
    			boolean burykings = booleanRead(piledata, "burykings", false);
    			
    			//And the display attributes
    			int displayx = intRead(piledata, "displayx", 32);
    			int displayy = intRead(piledata, "displayy", 32);
    			boolean straightstack = booleanRead(piledata, "straightstack", false);

    			//And add a pile based on these variables to the state.
    			state.addPile(new Pile(id, capacity, order, uniquesuit, alternatecolours, samesuit, removable, addable, isfoundation, stocknumber, cardstodeal, isstock, stockcapacity, startvalue, oneaway, oneawaynowrap, fourteens, displayx, displayy, straightstack, backwards));
    			
    			//Store BuryKings properties separately - this is not needed in-game.
    			if (burykings) {
    				pilestoburykings.add(new Integer(id));
    			}
    		}
    		
    		//Interpret the recommended tactics for solving the game.
    		LinkedList tacticslist = new LinkedList();
    		NodeList tactics = document.getElementsByTagName("tactic");
    		for (int i = 0; i < tactics.getLength(); i++) {
    			NamedNodeMap tacticdata = (tactics.item(i)).getAttributes();
    			String tactic = tacticdata.getNamedItem("name").getNodeValue();
    			tacticslist.add(tactic);
    		}
    		//Convert the linked list to an array of strings and add it to the state.
    		String[] gametactics = new String[tacticslist.size()];
    		for (int i = 0; i < tacticslist.size(); i++) {
    			gametactics[i] = (String)tacticslist.get(i);
    		}
    		
    		state.gametactics = gametactics;
    		
    		//Now set up the pack to be dealt into the tableau.
    		LinkedList pack = makedecks(deckstouse);
    		
    		NodeList fixedcards = document.getElementsByTagName("card");
    		LinkedList dealtcards = new LinkedList();
    		
    		//Deal the fixed cards into the specified piles.
    		for (int i = 0; i < fixedcards.getLength(); i++) {
    			NamedNodeMap carddata = (fixedcards.item(i)).getAttributes();
    			int pile = Integer.parseInt(carddata.getNamedItem("pile").getNodeValue());
    			int suit = Integer.parseInt(carddata.getNamedItem("suit").getNodeValue());
    			int value = Integer.parseInt(carddata.getNamedItem("value").getNodeValue());
    			((Pile)state.piles.get(pile)).dealCard(new Card(suit, value));
    			//Record that this card already exists in the game.
    			dealtcards.add(new Card(suit, value));
    		}
    		
    		//Mark any cards that have already been dealt.
    		int j = 0;
    		while (j < dealtcards.size()) {
    			//Get the card to find.
    			Card cardtofind = (Card)dealtcards.get(j);
    			
    			//Now loop through the pack until it is found.
    			int i = 0;
    			while (!((Card)pack.get(i)).equals(cardtofind)) {
    				i++;
    			}
    			//The card has been reached. Mark it.
    			Card foundcard = (Card)pack.get(i);
    			foundcard.value = 0;
    			pack.set(i, foundcard);
    			
    			//And move to the next dealt card.
    			j++;
    		}
    		
    		//Now remove the marked cards.
    		for (int i = 0; i < pack.size(); i++) {
    			if (((Card)pack.get(i)).value == 0) {
    				pack.remove(i);
    				i--;
    			}
    		}
    		
    		//Shuffle the remaining cards.
    		if (pack.size() > 0) {
    			pack = shufflepack(pack, 1000);

    		
    			//Find the stock piles.
    			//LinkedList stocks = new LinkedList();
    			for (int i = 0; i < state.piles.size(); i++) {
    				if (((Pile)state.piles.get(i)).isstock) {
    					//Fill the stock.
    					Pile stock = ((Pile)state.piles.get(i));
    					j = 0;
    					while (stock.stackpointer < stock.contents.length) {
    						stock.dealCard((Card)pack.get(0));
    						pack.remove(0);
    					}
    					state.piles.set(i, stock);
    				}
    			}
    		}
    		
    		//Deal the rest of the cards into the piles that need dealt cards.
    		
    		//Get the dealable piles.
            LinkedList dealables = new LinkedList();
            for (int i = 0; i < state.piles.size(); i++) {
            	if (((Pile)state.piles.get(i)).cardstodeal > 0) {
            		dealables.add(new Integer(i));
            	}
            }

            //Now deal into them.
        	for (int i = 0; i < dealables.size(); i++) {
        		//Get the pile from the list of dealables.
        		int index = ((Integer)dealables.get(i)).intValue();
        		Pile dealingpile = (Pile)(state.piles.get(index));
        		//Deal into it until it has reached the required number.
        		for (j = 0; j < dealingpile.cardstodeal; j++) {
        			((Pile)(state.piles.get(index))).dealCard((Card)pack.remove(0));
        		}
        	}
        	
        	//Bury the kings, if necessary.
        	for (int i = 0; i < pilestoburykings.size(); i++) {
        		//Get the pile number.
        		int pilenumber = ((Integer)pilestoburykings.get(i)).intValue();
        		//Get the pile.
        		((Pile)state.piles.get(pilenumber)).buryKings();
        	}
        	
        	//Find the numbers of the foundation piles and store them in an array.
            int countfoundations = 0;
            
            for (int i = 0; i < state.piles.size(); i++) {
            	if (((Pile)state.piles.get(i)).isfoundation) {
            		countfoundations++;
            	}
            }
            int[] foundations = new int[countfoundations];
            int index = 0;
            for (int i = 0; i < state.piles.size(); i++) {
            	if (((Pile)state.piles.get(i)).isfoundation) {
            		foundations[index] = i;
            		index++;
            	}
            }
            state.foundations = foundations;
    	}
    	//Catch any exceptions generated (code from tutorial)
    	catch (SAXException sxe) {
    		// Error generated during parsing
    		Exception  x = sxe;
    		if (sxe.getException() != null)
    			x = sxe.getException();
    		x.printStackTrace();

    	} 
    	catch (ParserConfigurationException pce) {
    		// Parser with specified options can't be built
    		pce.printStackTrace();
        }
    	catch (IOException ioe) {
    		// I/O error
    		ioe.printStackTrace();
     	}
    	catch (IndexOutOfBoundsException ioobe) {
    		JOptionPane.showMessageDialog(null, "Too many cards required!\nCheck that the number of cards to deal totals the expected value.", "Too many cards", JOptionPane.ERROR_MESSAGE);
    		ioobe.printStackTrace();
    	}
    	state.sourcemove = null;
    	return (state);
	}
	
    /**
     * Sets up the initial set of cards to use - a number of packs specified by the file.
     * @return A LinkedList consisting of a multiple of 52 Card objects, 13 values of each of the 4 suits.
     */
	public LinkedList makedecks(int deckstogenerate) {
        LinkedList cards = new LinkedList();
        
        for (int decks = 0; decks < deckstogenerate; decks++) {
        
        	for (int i = 1; i < 14; i++) {
        		for (int j = 0; j < 4; j++) {
        			cards.add(new Card(j, i));
        		}
        	}
        }
        return (cards);
    }
	
    /**
     * Shuffles a given array of cards, by repeatedly swapping two positions chosen at random.
     * 
     * @param pack The array of cards to be shuffled.
     * @param shuffledepth The number of swaps to perform.
     * @return The shuffled pack.
     */
    public LinkedList shufflepack(LinkedList pack, int shuffledepth) {
        for (int i = 0; i < shuffledepth; i++) {
            int firstpos = (int)(Math.random() * pack.size());
            int secondpos = (int)(Math.random() * pack.size());
            Card tempcard = (Card)pack.get(firstpos);
            pack.set(firstpos, (Card)pack.get(secondpos));
            pack.set(secondpos, tempcard);
        }
        return (pack);
    }
    
    /**
     * Method to read a boolean property from the XML and to return it if it exists. If
     * an exception is encountered, the default value is returned instead.
     * 
     * @param piledata The node map from which to get data.
     * @param property The name of the property to search for.
     * @param defaultvalue The default value if the property is not found.
     * @return The boolean value of the property.
     */
    public boolean booleanRead(NamedNodeMap piledata, String property, boolean defaultvalue) {
    	try {
    		boolean value = (getBool(piledata.getNamedItem(property).getNodeValue()));
    		return (value);
    	}
    	catch (Exception e) {
    		return (defaultvalue);
    	}
    }
    
    /**
     * This method converts a String into a boolean, returning true if the contents of the String
     * are equal to "true", and false if the contents are anything else.
     * 
     * @param string The string to examine.
     * @return True if and only if the string is "true".
     */
	public boolean getBool(String string) {
		if (string.equals("true")) {
			return(true);
		}
		else {
			return(false);
		}
	}
    
    /**
     * Method to read an int property from the XML and to return it if it exists. If
     * an exception is encountered, the default value is returned instead.
     * 
     * @param piledata The node map from which to get data.
     * @param property The name of the property to search for.
     * @param defaultvalue The default value if the property is not found.
     * @return The value of the property, as an int.
     */
    public int intRead(NamedNodeMap piledata, String property, int defaultvalue) {
    	try {
    		int value = Integer.parseInt(piledata.getNamedItem(property).getNodeValue());
    		return (value);
    	}
    	catch (Exception e) {
    		return (defaultvalue);
    	}
    }
}
