/*
 * Pile.java
 *
 * Created on 19 November 2005, 13:12
 */

package cs4099;

/**
 * A class that represents a pile in a Solitaire game. Piles hold cards, and have specific
 * rules as to the cards that can be placed on them at any time.
 *
 * @author David Newton
 * @see Card
 * @see State
 */

public class Pile {
    
    //Attributes dealing with the contents
    public int id;
    public Card[] contents;
    public int stackpointer = 0; //Stores where the next card will be added.
    
    //Attributes that store the constraints on this pile
    public boolean alternatecolours; //Cards must be placed in alternating colours.
    public int order; //The decrement/increment for new cards on the pile (0 if none)
    public boolean samesuit; //Cards must all be of the same suit (used mostly for Foundation/goal piles).
    public int uniquesuit; //Cards must be of a certain suit.
    public boolean isstock; //This pile will deal to all tableau piles when moved from.
    public int stockcapacity; //The number of cards to have in the stock.
    public boolean isfoundation; //This is a goal pile.
    public int startvalue; //The pile must start with this value. 0 if unrestricted.
    public boolean oneaway; //New cards must be one away from the top card.
    public boolean oneawaynowrap; //The same, with no wrapping between A and K.
    public int stocknumber; //The pile should be dealt to at the start of the game and by stock piles.
    public int cardstodeal; //The number of cards to deal to the pile at the beginning of the game.
    public boolean fourteens; //The cards added to the pile must hit fourteen every two adds.
    
    public boolean removable; //Cards can be removed from this pile.
    public boolean addable; //Cards can be added to this pile.
    
    //Attributes for visual display
    public int displayx;
    public int displayy;
    public boolean straightstack; //Set to "true" if the pile should be displayed showing only the top card.
    public boolean backwards; //If true, the pile will be displayed going upwards instead of downwards.
    
    /**
     * 
     * Creates a new instance of Pile.
     * 
     * @param id The ID number of the pile.
     * @param capacity The pile's capacity (length of the array used to hold cards)
     * @param order The numerical order in which cards must be placed - 1 specifies building up, -1 to build down.
     * @param uniquesuit Should be set to 0-3 if the pile should only accept one suit, -1 otherwise.
     * @param alternatecolours Set to true if cards must be placed on the pile in alternating colours.
     * @param samesuit Set to true if cards added to the pile must be of the same suit as the card below.
     * @param removable Set to true if cards can be removed from this pile (unless the pile is a stock).
     * @param addable Set to true if cards can be added from this pile.
     * @param isfoundation Set to true if this pile is to be regarded as a goal pile.
     * @param stocknumber The number of the stock that this pile is filled from (-1 if none).
     * @param cardstodeal The number of random cards to deal to this pile at the start of the game.
     * @param isstock Set to true if this pile should behave as a stock pile.
     * @param stockcapacity The number of cards to deal from this pile (if a stock) at one time.
     * @param startvalue The value that a card must be to fill this pile if it is empty (0 if any).
     * @param oneaway Set to true if cards with adjacent values must be added to this pile (wrapping between A-K allowed).
     * @param oneawaynowrap Set to true if cards with adjacent values must be added to this pile (wrapping between A-K disallowed).
     * @param fourteens Set to true if this pile must accept cards in pairs that total fourteen numerically.
     * @param displayx The X position for this pile on the desktop.
     * @param displayy The Y position for this pile on the desktop.
     * @param straightstack Set to true if only the top card is to be displayed.
     * @param backwards Set to true if this pile should be drawn upwards instead of downwards.
     */
    public Pile(int id, int capacity, int order, int uniquesuit, boolean alternatecolours, boolean samesuit, boolean removable, boolean addable, boolean isfoundation, int stocknumber, int cardstodeal, boolean isstock, int stockcapacity, int startvalue, boolean oneaway, boolean oneawaynowrap, boolean fourteens, int displayx, int displayy, boolean straightstack, boolean backwards) {
        
        this.id = id;
        this.contents = new Card[capacity];
        this.stackpointer = 0;
        this.alternatecolours = alternatecolours;
        this.samesuit = samesuit;
        this.uniquesuit = uniquesuit;
        this.removable = removable;
        this.addable = addable;
        this.isfoundation = isfoundation;
        this.stocknumber = stocknumber;
        this.cardstodeal = cardstodeal;
        this.isstock = isstock;
        this.stockcapacity = stockcapacity;
        this.order = order;
        this.startvalue = startvalue;
        this.oneaway = oneaway;
        this.oneawaynowrap = oneawaynowrap;
        this.fourteens = fourteens;
        this.displayx = displayx;
        this.displayy = displayy;
        this.straightstack = straightstack;
        this.backwards = backwards;
    }
    
    /**
     * Returns a String representation of this pile for output.
     * @return The string representation of card suits and values, separated by tabs, or [Empty] if empty.
     */
    public String toString() {
        
        String strung = "";
        
        //Indicate an empty pile if the stack pointer is at 0
        if (0 == stackpointer) {
            return("[Empty]");
        }
        
        else {
            for (int i = 0; i < stackpointer; i++) {
                strung += (contents[i].toString() + "\t");
            }
        }
        return(strung);
    }
    
    /**
     * Adds a card on to a pile, if it does not go against the constraints of the pile.
     * @return True if the addition was successful.
     */
    public boolean addCard(Card card) {
        
        if (this.canAddCard(card).equals("")) {
            //Add the card to the pile, incrementing the pointer.
            this.contents[stackpointer] = card;
            this.stackpointer++;
            return true;
        }
        else {
            return false;
        }
    }
    
    /**
     * Forces adding a card to a pile - used in dealing the cards rather than playing them. The pile cannot exceed its capacity.
     * @return True if the deal was successful.
     */
    public boolean dealCard(Card card) {
        if ((this.stackpointer) >= this.contents.length) {
            //System.out.println(card.toString() + " cannot be dealt to " + contents[stackpointer-1].toString() + " in pile " + this.id + " - the pile is full");
            //System.out.println(card.toString() + " NOT dealt to pile " + this.id);
            return false;
        }
        else {
            this.contents[stackpointer] = card;
            this.stackpointer++;
            //System.out.println(card.toString() + " dealt to pile " + this.id);
            return true;
        }
    }
    
    /**
     * Determines whether a card can be added to the pile. For the purposes of error output, the
     * method has to return a String that explains the state of the move - this string will be empty
     * if the move is valid.
     * 
     * @param card The card to attempt to add to the pile.
     * @return The error message, or an empty string if the move is valid.
     */
    public String canAddCard(Card card) {
        //Can't add...
    	//If the pile cannot have cards added to it.
    	if ((this.addable == false)) {
    		return ("Cards cannot be added to this pile.");
    	}
        //If the card will overfill the pile.
        else if ((this.stackpointer) >= this.contents.length) {
            return ("This move would overfill the pile, which has a limit of " + this.contents.length);
        }
        //If the card is of a suit not accepted by this pile.
        else if ((this.uniquesuit > -1) && card.suit != this.uniquesuit) {
        	return ("The card does not match the pile's suit");
        }
        //If the pile must have alternate colours, and there are other cards in the pile, and the card colours match
        else if (this.alternatecolours && (stackpointer > 0) && ((contents[stackpointer-1].suit % 2) == (card.suit % 2))) {
            return ("This pile must have alternate-coloured cards placed on it.");
        }
        //If the cards must be in order, and there are other cards in the pile, and the card's value does not match the order
        else if ((this.order != 0) && (stackpointer > 0) && ((contents[stackpointer-1].value + this.order) != (card.value))) {
            return ("The card does not follow the order of the pile: " + this.order);
        }
        //If the cards have to be the same suit, and there are other cards in the pile, and they do not match.
        else if ((this.samesuit) && (stackpointer > 0) && ((contents[stackpointer-1].suit) != (card.suit))) {
            return ("Cards on this pile must be of the same suit");
        }
        //If the card is the first and does not match the start value (if any).
        else if ((this.startvalue != 0) && (stackpointer == 0) && (card.value != this.startvalue)) {
        	return ("The pile must start with a card of value " + this.startvalue);
        }
        else if ((this.fourteens && this.stackpointer % 2 == 1 && this.getTopCard().value + card.value != 14)) {
        	return ("The pile's values must total a multiple of 14.");
        }
        //If the cards must one away by value, and there are other cards in the pile, and the card's value is not one away (or wrapped around)
        else if (this.oneaway && (stackpointer > 0) && !( (1 == (Math.abs(contents[stackpointer-1].value - card.value))) || (contents[stackpointer-1].value == 1 && card.value == 13) || (card.value == 1 && contents[stackpointer-1].value == 13))) {
        	return ("The card's value must be one away from the card below it (A and K are adjacent).");
        }
        //If the cards must one away by value, and there are other cards in the pile, and the card's value is not one away (no wrapping)
        else if (this.oneawaynowrap && (stackpointer > 0) && !(1 == (Math.abs(contents[stackpointer-1].value - card.value)))) {
        	return ("The card's value must be one away from the card below it.");
        }
        //But if all these conditions have been passed, the card can be added.
        else {
            return "";
        }
    }
    
    /**
     * Returns the topmost card, or a null if the topmost card cannot be removed.
     * @return The topmost card, or null.
     */
    public Card removeCard() {
        //Don't remove if cards cannot be taken from here.
        if (this.canRemoveCard() || this.isstock) {
        	this.stackpointer--;
            //System.out.println(this.contents[stackpointer].toString() + " removed from pile " + this.id);
            return (this.contents[stackpointer]);
        }
        else {
            System.err.println("Card cannot be removed from pile " + this.id);
            return (null);
        }
    }
    
    /**
     * Verifies whether the top card can be removed, without actually removing it.
     * @return True if it can be removed, False otherwise.
     */
    public boolean canRemoveCard() {
    	if ((this.removable && this.stackpointer > 0)) {
            return (true);
        }
    	else {
    		return (false);
    	}
    }
    
    /**
     * Returns the top card of this pile - a shortcut for getting its suit and value.
     * @return The top card, or null if there are no cards on the pile.
     */
    public Card getTopCard() {
    	if (stackpointer == 0) {
    		return (null);
    	}
    	else {
    		return (this.contents[stackpointer-1]);
    	}
    }
    
    /**
     * Examines whether a pile's properties are equivalent to another, including the suits and values of the cards in them.
     * @param pile The pile to compare against.
     * @return True if the piles are identical by properties, False if not.
     */
    public boolean isequivalent(Pile pile) {
    	
    	if (this.id == pile.id) {return true;}
    	else if (this.order != pile.order) {return false;}
    	else if (this.uniquesuit != pile.uniquesuit) {return false;}
    	else if (this.alternatecolours != pile.alternatecolours) {return false;}
    	else if (this.samesuit != pile.samesuit) {return false;}
    	else if (this.addable != pile.addable) {return false;}
    	else if (this.removable != pile.removable) {return false;}
    	else if (this.isfoundation != pile.isfoundation) {return false;}
    	else if (this.fourteens != pile.fourteens) {return false;}
    	else if (this.isstock != pile.fourteens) {return false;}
    	else if (this.stockcapacity != pile.stockcapacity) {return false;}
    	else if (this.stocknumber != pile.stocknumber) {return false;}
    	else if (this.startvalue != pile.startvalue) {return false;}
    	else if (this.oneaway != pile.oneaway) {return false;}
    	else if (this.oneawaynowrap != pile.oneawaynowrap) {return false;}
    	else if (!this.equalsByContents(pile)) {return false;}
    	else {return true;}
    	
    }
    
    /**
     * Returns information on whether any of a set of cards exist in this pile.
     * The return type is int rather than boolean, as this also provides the position that the
     * card has in the pile relative to the top card.
     * 
     * @param cards The array of cards to find.
     * @return The number of cards below the surface at which a card is found, or -1 if there was no match.
     */
    public int contains(Card[] cards) {
    	
    	int carddepth = -1;
    	
    	for (int i = 0; i < this.stackpointer-1; i++) {
    		for (int j = 0; j < cards.length; j++) {
    			if (cards[j].equals(this.contents[i])) {
    				//System.out.println(this.id + " contains required card");
    				carddepth = this.stackpointer - i;
    			}
    		}
    	}
    	return (carddepth);
    }
    
    /**
     * Returns "true" if this pile is found to be "greater than" another, based on its attributes and contents but not its ID number.
     * @param pile The pile to compare to
     * @return 1 if this pile is greater, -1 if the passed pile is greater, 0 if they are equivalent
     */
    public int compareTo(Pile pile) {
    	//Go through all attributes first.
    	if (this.contents.length > pile.contents.length) {
    		return 1;
    	}
    	else if (this.contents.length < pile.contents.length) {
    		return -1;
    	}
    	
    	else if (this.order > pile.order) {
    		return 1;
    	}
    	else if (this.order < pile.order) {
    		return -1;
    	}
    	
    	else if (this.uniquesuit > pile.uniquesuit) {
    		return 1;
    	}
    	else if (this.uniquesuit < pile.uniquesuit) {
    		return -1;
    	}
    	
    	else if (this.alternatecolours == true && pile.alternatecolours == false) {
    		return 1;
    	}
    	else if (this.alternatecolours == false && pile.alternatecolours == true) {
    		return -1;
    	}
    	
    	else if (this.samesuit == true && pile.samesuit == false) {
    		return 1;
    	}
    	else if (this.samesuit == false && pile.samesuit == true) {
    		return -1;
    	}
    	
    	else if (this.removable == true && pile.removable == false) {
    		return 1;
    	}
    	else if (this.removable == false && pile.removable == true) {
    		return -1;
    	}
    	
    	else if (this.addable == true && pile.addable == false) {
    		return 1;
    	}
    	else if (this.addable == false && pile.addable == true) {
    		return -1;
    	}
    	
    	else if (this.isfoundation == true && pile.isfoundation == false) {
    		return 1;
    	}
    	else if (this.isfoundation == false && pile.isfoundation == true) {
    		return -1;
    	}
    	
    	else if (this.fourteens == true && pile.fourteens == false) {
    		return 1;
    	}
    	else if (this.fourteens == false && pile.fourteens == true) {
    		return -1;
    	}
    	
    	else if (this.isstock == true && pile.isstock == false) {
    		return 1;
    	}
    	else if (this.isstock == false && pile.isstock == true) {
    		return -1;
    	}
    	
    	else if (this.stockcapacity > pile.stockcapacity) {
    		return 1;
    	}
    	else if (this.stockcapacity < pile.stockcapacity) {
    		return -1;
    	}
    	
    	else if (this.stocknumber > pile.stocknumber) {
    		return 1;
    	}
    	else if (this.stocknumber < pile.stocknumber) {
    		return -1;
    	}
    	
    	else if (this.startvalue > pile.startvalue) {
    		return 1;
    	}
    	else if (this.startvalue < pile.startvalue) {
    		return -1;
    	}
    	
    	else if (this.oneaway == true && pile.oneaway == false) {
    		return 1;
    	}
    	else if (this.oneaway == false && pile.oneaway == true) {
    		return -1;
    	}
    	
    	else if (this.oneawaynowrap == true && pile.oneawaynowrap == false) {
    		return 1;
    	}
    	else if (this.oneawaynowrap == false && pile.oneawaynowrap == true) {
    		return -1;
    	}
    	
    	//If all these are the same, look at the first card in each pile - no need to go further.
    	
    	Card thiscard = this.getTopCard();
    	Card pilecard = pile.getTopCard();
    	
    	if (thiscard == null && pilecard == null) {
    		return 0;
    	}
    	else if (thiscard == null && pilecard != null) {
    		return -1;
    	}
    	else if (pilecard == null && thiscard != null) {
    		return 1;
    	}
    	else if (thiscard.value > pilecard.value) {
    		return 1;
    	}
    	else if (thiscard.value < pilecard.value) {
    		return -1;
    	}
    	else if (thiscard.suit > pilecard.suit) {
    		return 1;
    	}
    	else if (thiscard.suit < pilecard.suit) {
    		return -1;
    	}
    	return 0;
    }
    
    /**
     * A method to compare the contents of this pile with another, and to return "true" if the contents are equal.
     * @param otherpile The pile to compare to.
     * @return "True" if the contents are equal, "False" otherwise.
     */
    public boolean equalsByContents(Pile otherpile) {
    	Card[] thiscontents = this.contents;
    	Card[] othercontents = otherpile.contents;
    	if (this.stackpointer != otherpile.stackpointer+1) {
    		return (false);
    	}
    	else {
    		//Iterate through and compare cards.
    		for (int i = 0; i < otherpile.stackpointer; i++) {
    			if (!thiscontents[i].equals(othercontents[i])) {
    				return (false);
    			}
    		}
    	}
    	return (true);
    }
    
    /**
     * Buries the Kings on this pile (sends all cards with value 13 to the beginning).
     * This is used in the dealing phase of some games.
     */
    public void buryKings() {
    	//Loop through the pile and send the kings to the start.
    	Card[] newcontents = new Card[this.contents.length];
    	int newstackpointer = 0;
    	//Get the Kings first.
    	for (int i = 0; i < this.stackpointer; i++) {
    		if (this.contents[i].value == 13) {
    			newcontents[newstackpointer] = new Card(this.contents[i].suit, this.contents[i].value);
    			newstackpointer++;
    		}
    	}
    	//Now copy in the non-Kings.
    	for (int i = 0; i < this.stackpointer; i++) {
    		if (this.contents[i].value != 13) {
    			newcontents[newstackpointer] = new Card(this.contents[i].suit, this.contents[i].value);
    			newstackpointer++;
    		}
    	}
    	//Replace the contents of this pile.
    	this.contents = newcontents;
    }
}
