package cs4099;

import javax.swing.*;
import java.util.*;

public class ScriptedSolver extends JFrame {
	
	public int statedepth = 0;
	public int depthlimit = 0;
	public int examinedstates = 0;
	public int enqueuedstates = 0;
	
	public boolean cancelled = false;
	
	private JLabel labeldepth = new JLabel();
	private JLabel labelexamined = new JLabel();
	private JLabel labelenqueued = new JLabel();
	private JLabel labelbest = new JLabel();
	
	public State rootstate;
	private State currentstate;
	
	public State beststate;
	public Move[] solutionpath = {};
	
	/**
	 * Initializes a new Solver.
	 * 
	 * @param rootstate The state to solve from
	 * @param depthlimit The maximum depth to search
	 */
	public ScriptedSolver(State rootstate, int depthlimit) {
		this.rootstate = rootstate;
		this.beststate = rootstate;
		this.depthlimit = depthlimit;
	}
	
	/**
	 * Initializes the visual components of this solver.
	 */
	private void initComponents() {
		
//		Essentials
		getContentPane().setLayout(new AbsoluteLayout());
        setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        setResizable(false);
		
		getContentPane().add(labeldepth, new AbsoluteConstraints(30, 30, 200, 20));
		getContentPane().add(labelexamined, new AbsoluteConstraints(30, 50, 200, 20));
		getContentPane().add(labelenqueued, new AbsoluteConstraints(30, 70, 200, 20));
		getContentPane().add(labelbest, new AbsoluteConstraints(30, 90, 200, 20));
		
		this.setSize(300, 150);
		this.setTitle("Solver - Running");
		this.setVisible(true);
	}
	
	/**
	 * Runs the solver thread.
	 */
	public String solvefromstate() {
		
		initComponents();
		String returnstring = "";
		try {
			returnstring = solvegame(rootstate);
		}
		
		catch (OutOfMemoryError e) {
			returnstring = "Out of memory at " + beststate.countCardsInFoundations();
		}
		return returnstring;
	}
	
	/**
	 * Updates the on-screen counters to show progress.
	 *
	 */
	private void updateCounters() {
		labeldepth.setText("Depth: " + statedepth);
		labelexamined.setText("Examined states: " + examinedstates);
		labelenqueued.setText("Enqueued states: " + enqueuedstates);
		labelbest.setText("Best state: " + beststate.countCardsInFoundations());
		getContentPane().paintAll(getContentPane().getGraphics());
		this.setTitle("Solver - Running");
	}
	
	/**
	 * Solves the game by expanding a tree of states from the root.
	 * 
	 * @param rootstate The state from which to begin.
	 */
	private String solvegame(State rootstate) {
    	
    	Hashtable hashstore = new Hashtable(65536);
        currentstate = rootstate;
        currentstate.requiredcards = currentstate.findrequiredcards();
        StateStack queue = new StateStack();
        
    	//Add the root state to the queue.
        queue.add(currentstate);
        
        /*
         * MAIN SOLVING LOOP
         */
        
    	//While the game has not been solved, found to be impossible or cancelled...
        
        while(!queue.isEmpty() && !gameSolved(currentstate) && !this.cancelled) {

        	//Dequeue the next state to examine.
        	currentstate = queue.get();
        	examinedstates++;
        	statedepth = currentstate.depth;
        	enqueuedstates = queue.list.size();

        	updateCounters();
        	
        	//Give the user a chance to interrupt the thread.
        	try {
        		Thread.sleep(1);
        	}
        	catch (Exception e) {
        		this.cancelled = true;
        	}

        	Move[] validmoves = currentstate.findValidMoves();
        	
        	for (int i = 0; i < validmoves.length; i++) {
        		
        		//Copy the current state and perform the move on it to generate a new one.
        		State newstate = currentstate.copy();
        		newstate.performMove(validmoves[i].sourcepileno, validmoves[i].targetpileno);
        		newstate.sourcemove = validmoves[i];
        		newstate.parent = currentstate;
        		newstate.depth = currentstate.depth+1;
        		newstate.requiredcards = newstate.findrequiredcards();
        		
        		//Only enqueue the state if an identical state has not been reached on this branch.
        		State sortedstate = newstate.getSortedState();
        		
        		//If a duplicate is not found, store the sorted state in the map and the original state in the queue.
        		if (!hashstore.containsValue(new Integer(sortedstate.hashCode())) && (sortedstate.depth < depthlimit)) {
        			
        			hashstore.put(sortedstate, new Integer(sortedstate.hashCode()));
        			queue.add(newstate);
        			
        			//If the new state is better than the old best state, replace it.
        			if (newstate.countCardsInFoundations() > beststate.countCardsInFoundations()) {
        				beststate = newstate.copy();
        			}
        		}
        		else {
        			
        			//Ignore the new state if it matches one already in the hashstore.
        			if (hashstore.get(sortedstate) != null) {
        				hashstore.put(sortedstate, new Integer(sortedstate.hashCode()));
        			}
        		}
        		
        	}
        }
        
        //Report the outcome of the solver.
        if (gameSolved(currentstate)) {
        	return("Solved in " + statedepth);
        }
        else if (!this.cancelled){
        	return("Impossible after " + examinedstates);
        }
        else {
        	return("Cancelled");
        }
    }
	
	/**
	 * Returns true if a game has been solved (all cards are in piles marked as foundations).
	 * @param state The state to check.
	 * @return True if the game has been solved, False otherwise.
	 */
	private boolean gameSolved(State state) {
		Pile pile;
		for (int i = 0; i < currentstate.piles.size(); i++) {
			pile = (Pile)(currentstate.piles.get(i));
			if (!pile.isfoundation && pile.stackpointer != 0) {
				return(false);
			}
		}
		return (true);
	}
	
	/**
	 * Gets the path that was taken to arrive at a certain state by iterating through its parents.
	 * 
	 * @param state The state from which to retrieve the path
	 * @return The array of moves that led to the specified state.
	 */
	private Move[] getPath(State state) {
    	//Fill the solution path from the last state.
    	Move[] moves = new Move[state.depth];
    	
    	for (int i = moves.length-1; i >= 0; i--) {
    		moves[i] = state.sourcemove;
    		state = state.parent;
    	}
    	return (moves);
    }
}
