package cs4099;

import javax.swing.*;
import java.util.*;

/**
 * The Solver class takes in a root state and expands a tree of possible future states. Priority
 * is given to moves that match the set of tactics given in the state. It runs as a thread, so can
 * be cancelled during the process.
 * 
 * The solver will only search to the maximum depth specified - if this is hit, it will backtrack
 * and attempt another route.
 * 
 * @author David Newton
 * @see State
 * @see StateStack
 */
public class Solver extends JFrame implements Runnable {
	
	private int statedepth = 0;
	private int depthlimit = 0;
	private int examinedstates = 0;
	private int enqueuedstates = 0;
	private boolean limitreached = false;
	
	public boolean cancelled = false;
	
	private JLabel labeldepth = new JLabel();
	private JLabel labelexamined = new JLabel();
	private JLabel labelenqueued = new JLabel();
	private JLabel labelbest = new JLabel();
	
	private State rootstate;
	private State currentstate;
	
	private State beststate;
	public Move[] solutionpath = {};
	
	/**
	 * Initializes a new Solver.
	 * 
	 * @param rootstate The state to solve from
	 * @param depthlimit The maximum depth to search
	 */
	public Solver(State rootstate, int depthlimit) {
		this.rootstate = rootstate;
		this.beststate = rootstate;
		this.depthlimit = depthlimit;
	}
	
	/**
	 * Initializes the visual components of this solver.
	 */
	private void initComponents() {
		
//		Essentials
		getContentPane().setLayout(new AbsoluteLayout());
        setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        setResizable(false);
		
		getContentPane().add(labeldepth, new AbsoluteConstraints(30, 30, 200, 20));
		getContentPane().add(labelexamined, new AbsoluteConstraints(30, 50, 200, 20));
		getContentPane().add(labelenqueued, new AbsoluteConstraints(30, 70, 200, 20));
		getContentPane().add(labelbest, new AbsoluteConstraints(30, 90, 200, 20));
		
		this.setSize(300, 150);
		this.setTitle("Solver - Running");
		this.setVisible(true);
	}
	
	/**
	 * Runs the solver thread.
	 */
	public void run() {
		
		initComponents();
		try {
			solveGame(rootstate);
		}
		
		catch (OutOfMemoryError e) {
			String windowstring = "Out of memory.\nUse the Retrieve button to get the closest path.";
			if (limitreached) { windowstring += "\nThe depth limit was reached."; limitreached = false; }
			JOptionPane.showMessageDialog(null, windowstring, "Too many states", JOptionPane.ERROR_MESSAGE);
			solutionpath = getPath(beststate);
			this.setTitle("Solver - Stopped");
		}
	}
	
	/**
	 * Updates the on-screen counters to show progress.
	 *
	 */
	private void updateCounters() {
		labeldepth.setText("Depth: " + statedepth);
		labelexamined.setText("Examined states: " + examinedstates);
		labelenqueued.setText("Enqueued states: " + enqueuedstates);
		labelbest.setText("Best state: " + beststate.countCardsInFoundations());
		getContentPane().paintAll(getContentPane().getGraphics());
		this.setTitle("Solver - Running");
	}
	
	/**
	 * Solves the game by expanding a tree of states from the root.
	 * 
	 * @param rootstate The state from which to begin.
	 */
	private void solveGame(State rootstate) {
    	
    	Hashtable hashstore = new Hashtable(65536);
        currentstate = rootstate;
        currentstate.requiredcards = currentstate.findrequiredcards();
        StateStack queue = new StateStack();
        
    	//Add the root state to the queue.
        queue.add(currentstate);
        
        /*
         * MAIN SOLVING LOOP
         */
        
    	//While the game has not been solved, found to be impossible or cancelled...
        
        while(!queue.isEmpty() && !gameSolved(currentstate) && !this.cancelled) {

        	//Dequeue the next state to examine.
        	currentstate = queue.get();
        	examinedstates++;
        	statedepth = currentstate.depth;
        	enqueuedstates = queue.list.size();

        	updateCounters();
        	
        	//Give the user a chance to interrupt the thread.
        	try {
        		Thread.sleep(1);
        	}
        	catch (Exception e) {
        		this.cancelled = true;
        	}

        	Move[] validmoves = currentstate.findValidMoves();
        	
        	for (int i = 0; i < validmoves.length; i++) {
        		
        		//Copy the current state and perform the move on it to generate a new one.
        		State newstate = currentstate.copy();
        		newstate.performMove(validmoves[i].sourcepileno, validmoves[i].targetpileno);
        		newstate.sourcemove = validmoves[i];
        		newstate.parent = currentstate;
        		newstate.depth = currentstate.depth+1;
        		newstate.requiredcards = newstate.findrequiredcards();
        		
        		//Only enqueue the state if an identical state has not been reached on this branch.
        		State sortedstate = newstate.getSortedState();
        		
        		//If a duplicate is not found, store the sorted state in the map and the original state in the queue.
        		if (!hashstore.containsValue(new Integer(sortedstate.hashCode())) && (sortedstate.depth < depthlimit)) {
        			
        			hashstore.put(sortedstate, new Integer(sortedstate.hashCode()));
        			queue.add(newstate);
        			
        			//If the new state is better than the old best state, replace it.
        			if (newstate.countCardsInFoundations() > beststate.countCardsInFoundations()) {
        				beststate = newstate.copy();
        			}
        		}
        		else if (sortedstate.depth < depthlimit) {
        			
        			//Ignore the new state if it matches one already in the hashstore.
        			if (hashstore.get(sortedstate) != null) {
        				hashstore.put(sortedstate, new Integer(sortedstate.hashCode()));
        			}
        		}
        		else {
        			limitreached = true;
        		}
        		
        	}
        }
        
        //Report the outcome of the solver.
        if (gameSolved(currentstate)) {
        	String windowstring = "Found the solution in " + statedepth + " moves\nUse the Retrieve button to put this into the solution list.";
        	if (limitreached) { windowstring += "\nThe depth limit was reached."; limitreached = false; }
        	JOptionPane.showMessageDialog(null, windowstring, "Success", JOptionPane.INFORMATION_MESSAGE);
        	solutionpath = getPath(currentstate);
        	this.setTitle("Solver - Complete");
        }
        else if (!this.cancelled){
        	String windowstring = "";
        	if (limitreached) { windowstring = "No possible solution found within the given depth.\nClick Retrieve to get the path closest to the solution.\nA solution may be found if the depth limit is increased."; limitreached = false; }
        	else { windowstring = "No possible solution found.\nClick Retrieve to get the path closest to the solution."; }
        	JOptionPane.showMessageDialog(null, windowstring, "Failure", JOptionPane.ERROR_MESSAGE);
        	solutionpath = getPath(beststate);
        	this.setTitle("Solver - Stopped");
        }
        else {
        	solutionpath = getPath(beststate);
        	this.setTitle("Solver - Cancelled");
        }
    }
	
	/**
	 * Returns true if a game has been solved (all cards are in piles marked as foundations).
	 * @param state The state to check.
	 * @return True if the game has been solved, False otherwise.
	 */
	private boolean gameSolved(State state) {
		Pile pile;
		for (int i = 0; i < currentstate.piles.size(); i++) {
			pile = (Pile)(currentstate.piles.get(i));
			if (!pile.isfoundation && pile.stackpointer != 0) {
				return(false);
			}
		}
		return (true);
	}
	
	/**
	 * Gets the path that was taken to arrive at a certain state by iterating through its parents.
	 * 
	 * @param state The state from which to retrieve the path
	 * @return The array of moves that led to the specified state.
	 */
	private Move[] getPath(State state) {
    	//Fill the solution path from the last state.
    	Move[] moves = new Move[state.depth];
    	
    	for (int i = moves.length-1; i >= 0; i--) {
    		moves[i] = state.sourcemove;
    		state = state.parent;
    	}
    	return (moves);
    }
}
