
package cs4099;

import java.util.LinkedList;

/**
 * 
 * This class represents a collection of Piles - it holds a specific state in a Solitaire game.
 * 
 * @author David Newton
 * @see Pile
 */
public class State {
    
	public LinkedList piles;
    public State parent;
    public int depth;
    public Move sourcemove;
    public String[] gametactics;
    public int[] foundations;
    public Card[] requiredcards; //The cards required on the foundations from this state.
    public String gamename;
    
    /**
     * Creates a new State
     * @param parent The state from which the new state was created
     */
    public State(State parent) {
        this.piles = new LinkedList();
        this.parent = parent;
        this.depth = 0;
    }
    
    /**
     * Add a pile to this state (used in the interpretation stage).
     * @param pile The pile to add.
     */
    public void addPile(Pile pile) {
        piles.add(pile);
    }
    
    /**
     * Output this state to the command line in the form "[ID]: [Cards]".
     *
     */
    public void outputState() {
    	
    	for (int i = 0; i < this.piles.size(); i++) {
            System.out.println(((Pile)this.piles.get(i)).id + ":\t" + ((Pile)this.piles.get(i)).toString());
        }
    }
    
    /** 
     * Create an exact copy of this state, with duplicates of all piles and cards.
     * @return The copied state.
     */
    public State copy() {
    	State copiedstate = new State(null);
    	
    	for (int i = 0; i < this.piles.size(); i++) {
    		//Add new piles to the state.
    		Pile pile = (Pile)this.piles.get(i);
    		copiedstate.piles.add(new Pile(pile.id, pile.contents.length, pile.order, pile.uniquesuit, pile.alternatecolours, pile.samesuit, pile.removable, pile.addable, pile.isfoundation, pile.stocknumber, pile.cardstodeal, pile.isstock, pile.stockcapacity, pile.startvalue, pile.oneaway, pile.oneawaynowrap, pile.fourteens, pile.displayx, pile.displayy, pile.straightstack, pile.backwards));
    	}
    	//Now add the cards into these piles.
    	for (int i = 0; i < this.piles.size(); i++) {
    		Pile pile = (Pile)copiedstate.piles.get(i);
    		for (int j = 0; j < ((Pile)this.piles.get(i)).stackpointer; j++) {
    			//Copy the card into the new pile.
    			pile.dealCard(((Pile)this.piles.get(i)).contents[j]);
    		}
    	}
    	copiedstate.depth = this.depth;
    	copiedstate.parent = this.parent;
    	copiedstate.sourcemove = this.sourcemove;
    	copiedstate.foundations = this.foundations;
    	copiedstate.requiredcards = this.requiredcards;
    	copiedstate.gametactics = this.gametactics;
    	return copiedstate;
    }
    
    /**
     * Tests whether a move can be performed from this state, by considering two factors:
     * Whether the source pile can have its top card removed, and
     * whether the target pile can accept the top card from the first pile.
     * The return type is a String for error reporting reasons - it will be blank if the move
     * could be made successfully, and a description of the problem if it could not.
     * 
     * @param firstpile The source pile for the move
     * @param secondpile The target pile for the move
     * @return An empty string if the move can be performed, or an error message if it could not.
     */
    public String canPerformMove(Pile firstpile, Pile secondpile) {
    	
    	if (firstpile.getTopCard() != null) {

    		if (firstpile.isstock) {
    			if (firstpile.stackpointer > 0) {
    				//This move is OK.
    				return ("");
    			}
    			else {
    				return ("The stock has no cards");
    			}
    		}
    		String error = secondpile.canAddCard(firstpile.getTopCard());
    		if (firstpile.canRemoveCard() && error.equals("")) {
    			return ("");
    		}
    		else if (!firstpile.canRemoveCard()) {
    			return ("Card cannot be removed from pile " +firstpile.id);
    		}
    		else {
    			return (error);
    		}
    	}
    	else {
    		return ("The source pile has no cards.");
    	}
    }
    
    /**
     * Performs a move on the two piles corresponding to the passed values.
     * 
     * @param firstpileno The number of the source pile.
     * @param secondpileno The number of the target pile.
     */
    public void performMove(int firstpileno, int secondpileno) {
    	Pile firstpile = (Pile)piles.get(firstpileno);
    	Pile secondpile = (Pile)piles.get(secondpileno);
    	
    	//Special for stock piles.
    	if (firstpile.isstock) {
    		//Get the tableau. This is the collection of piles that need to be filled from this stock.
    		LinkedList tableau = new LinkedList();
            for (int i = 0; i < this.piles.size(); i++) {
            	if (((Pile)this.piles.get(i)).stocknumber == firstpile.id) {
            		tableau.add(new Integer(i));
            	}
            }
            //Now deal them in turn.
            for (int i = 0; i < firstpile.stockcapacity; i++) {
            	if (firstpile.stackpointer > 0) {
            		((Pile)this.piles.get(((Integer)tableau.get((i % tableau.size()))).intValue())).dealCard(firstpile.removeCard());
            	}
            }
    	}
    	
    	//Otherwise, moving the card is simple.
    	else {
    		Card card = firstpile.removeCard();
    		secondpile.addCard(card);
    	}
    	
    	//Write the modified piles back into the state.
    	this.piles.set(firstpileno, firstpile);
    	this.piles.set(secondpileno, secondpile);
    }
    
    /**
     * Method to test the state against another to check for duplicates.
     * @param newstate The state to compare with.
     * @return True if the states are equivalent, False if not.
     */
    public boolean equals(State newstate) {
    	
    	//Iterate through all the piles and cards in this state, comparing them to the new state.
    	
    	for (int pileno = 0; pileno < this.piles.size(); pileno++) {
    		
    		Pile firstpile = ((Pile)this.piles.get(pileno));
    		Pile secondpile = ((Pile)newstate.piles.get(pileno));
    		
    		//First test the size - if this is not the same, the state cannot be a duplicate
    		if (firstpile.stackpointer != secondpile.stackpointer) {
    			return (false);
    		}
    		
    		//Now iterate through the cards in the pile. If any are not the same, the state is not a duplicate.
    		for (int cardno = 0; cardno < firstpile.stackpointer-1; cardno++) {
    			if (!firstpile.contents[cardno].equals(secondpile.contents[cardno])) {
    				return (false);
    			}
    		}
    	}
    	
    	//If the program reaches here, no differences between the two states have been found.
    	return (true);
    }
    
    /**
     * Finds the cards required by the foundations in this state.
     * @return The array of required cards.
     */
    public Card[] findrequiredcards() {
    	LinkedList required = new LinkedList();
    	for (int i = 0; i < this.foundations.length; i++) {
    		//Get the foundation pile i.
    		Pile pile = ((Pile)this.piles.get(this.foundations[i]));
    		//Find the cards that will be accepted on this pile.
    		for (int suit = 0; suit < 4; suit++) {
    			for (int value = 1; value < 14; value++) {
    				Card card = new Card(suit, value);
    				if (pile.canAddCard(card).equals("")) {
    					//Add it to the list of acceptable cards.
    					required.add(card);
    				}
    			}
    		}
    	}
    	
    	Card[] cards = new Card[required.size()];
    	for (int i = 0; i < required.size(); i++) {
    		cards[i] = (Card)required.get(i);
    	}
    	return cards;
    }
    
    /**
     * Returns a new state with piles in order of properties, based on this state.
     * @return The sorted state.
     */
    public State getSortedState() {
    	State sortedstate = new State(this.parent);
    	State oldstate = this.copy();
    	//Copy the minimum pile from the old state into this one each time.
    	while (oldstate.piles.size() > 0) {
    		//Get the minimum.
    		Pile minimum = (Pile)oldstate.piles.get(0);
    		for (int i = 0; i < oldstate.piles.size(); i++) {
    			//If this pile is lower than the minimum, it is the new minimum.
    			//System.out.println(((Pile)oldstate.piles.get(i)).id + " " + minimum.id);
    			if (((Pile)oldstate.piles.get(i)).compareTo(minimum) < 0) {
    				
    				minimum = (Pile)oldstate.piles.get(i);
    				//System.out.println(((Pile)oldstate.piles.get(i)).id + " " + minimum.id);
    			}
    		}
    		//Add the minimum pile to the sorted state, and remove it from the old one.
    		sortedstate.addPile(minimum);
    		oldstate.piles.remove(minimum);
    	}
    	
    	//Now copy all the attributes.
    	sortedstate.depth = oldstate.depth;
    	sortedstate.foundations = oldstate.foundations;
    	//sortedstate.pathcost = oldstate.pathcost;
    	sortedstate.requiredcards = oldstate.requiredcards;
    	sortedstate.sourcemove = oldstate.sourcemove;
    	
    	return (sortedstate);
    }
    
    /**
     * Returns a hash code for this state.
     * @return The hash code.
     */
    public int hashCode() {
    	//Return a hash code for this object.
    	int hashcode = 0;
	
    	//Now get the values of the other piles.
    	for (int i = 0; i < this.piles.size(); i++) {
    		Pile pile = (Pile)this.piles.get(i);
    		
    		for (int j = 0; j < pile.stackpointer; j++) {
        		hashcode += (((14+j)*pile.contents[j].suit) + pile.contents[j].value);
        	}
    		
    		//And bitshift left.
    		hashcode = hashcode << 1;
    	}
    	
    	return hashcode;
    	
    }
    
    /**
     * Updates the foundations of this state - gets all piles with the "isFoundation" attribute set
     * to true, and adds their ID values into the foundations array for quick access.
     */
    public void updateFoundations() {
    	int countfoundations = 0;
        
        for (int i = 0; i < this.piles.size(); i++) {
        	if (((Pile)this.piles.get(i)).isfoundation) {
        		countfoundations++;
        	}
        }
        this.foundations = new int[countfoundations];
        int index = 0;
        for (int i = 0; i < this.piles.size(); i++) {
        	if (((Pile)this.piles.get(i)).isfoundation) {
        		this.foundations[index] = i;
        		index++;
        	}
        }
    }
    
    /**
     * Counts the number of cards in foundations.
     * @return The number of cards that are in foundations.
     */
    public int countCardsInFoundations() {
    	int cardsinfoundations = 0;
    	for (int i = 0; i < this.foundations.length; i++) {
    		Pile pile = (Pile)this.piles.get(this.foundations[i]);
    		cardsinfoundations += pile.stackpointer;
    	}
    	return (cardsinfoundations);
    }
    
    /**
     * Counts the number of cards in the entire state.
     * @return The number of cards found.
     */
    public int countCards() {
    	int cards = 0;
    	for (int i = 0; i < this.piles.size(); i++) {
    		Pile pile = (Pile)this.piles.get(i);
    		cards += pile.stackpointer;
    	}
    	return (cards);
    }

    /**
     * Finds and returns all the valid moves from this state.
     * @return The array of available moves.
     */
    public Move[] findValidMoves() {
    	
    	//Set up linked lists for array preparation.
    	LinkedList allmoves = new LinkedList();
    	
    	//Iterate through all piles, showing those where the top card of the first pile can be placed on the second pile.
    	for (int firstpileno = 0; firstpileno < this.piles.size(); firstpileno++) {
    		for (int secondpileno = 0; secondpileno < this.piles.size(); secondpileno++) {
    			
    			Pile firstpile = ((Pile)this.piles.get(firstpileno));
    			Pile secondpile = ((Pile)this.piles.get(secondpileno));
    			
//    			Special circumstances for stock piles.
    			if (firstpile.isstock) {
    				//Ignore this if it is empty, or the second pile is anything other than 0.
    				if (secondpile.id != 0 || firstpile.stackpointer == 0) {
    					continue;
    				}
    				//Add special move to the Move list.
    				allmoves.add(new Move(firstpileno, secondpileno, new Card(0, 0)));
    				continue;
    			}
    			
    			//Ignore the move if the pile numbers are equal, or the piles are equivalent.
    			if (firstpileno == secondpileno || firstpile.isequivalent(secondpile)) {
    				continue;
    			}
    			//Add it to the linked list if it meets the criteria.
    			else if (canPerformMove(firstpile, secondpile).equals("")) {
    				allmoves.add(new Move(firstpileno, secondpileno, firstpile.getTopCard()));
    			}
    		}
    	}
   	
    	//Now sort the linked list according to tactics.
    	allmoves = prioritize(allmoves);
    	
    	//Convert the linked list to an array and return it.
		Move[] movelist = new Move[allmoves.size()];
		int j = 0;
		for (int i = allmoves.size()-1; i >= 0; i--) {
			movelist[j] = (Move)allmoves.get(i);
			j++;
		}
    	
    	return(movelist);
    }
    
    /**
     * Method to sort a list of moves in priority according to the tactics of the game.
     * 
     * @param moves The list of moves to prioritize.
     * @return The same move list, but sorted into order of priority.
     */
    private LinkedList prioritize(LinkedList moves) {
    	
    	//The linked list in which to store the sorted moves.
    	LinkedList prioritizedmoves = new LinkedList();
    	
    	//Do the following for every tactic available.
    	for (int i = 0; i < this.gametactics.length; i++) {

    		String tactic = this.gametactics[i];
    		LinkedList matchedmoves = new LinkedList();
    		
    		//Iterate through all moves and remove the relevant ones.
    		for (int j = 0; j < moves.size(); j++) {
    			
    			Move examinedmove = (Move)moves.get(j);
    			
    			Pile sourcepile = ((Pile)(piles.get(examinedmove.sourcepileno)));
    			Pile targetpile = ((Pile)(piles.get(examinedmove.targetpileno)));
    			
    			//If the move involves moving to a foundation.
    			if (tactic.equals("movetofoundations") && !sourcepile.isstock) {
    				if (targetpile.isfoundation) {
    					matchedmoves.add(examinedmove);
    					moves.remove(j--);
    					continue;
    				}
    			}
    			
    			if (tactic.equals("uncoverrequired") && !sourcepile.isstock) {
    				examinedmove.priority = sourcepile.contains(this.requiredcards);
    				
    				//Prioritize the move if it brings a required card closer to the front.
    				//Ignore moves that have a distance of 0 - this is covered in the tactic above.
    				if (examinedmove.priority > 0) {
    					//Insert the move into the temporary list, by order of distance from front.
    					int index = 0;
    					while (matchedmoves.size() > index && ((Move)matchedmoves.get(index)).priority > examinedmove.priority) {
    						index++;
    					}
    					matchedmoves.add(index, examinedmove);
    					moves.remove(j--);
    					continue;
    				}
    			}
    			
    			if (tactic.equals("keepfreecells") && !sourcepile.isstock) {
    				//Prioritize moves that do not involve moving to an empty single cell.
    				if (targetpile.stackpointer > 0 && targetpile.contents.length == 1) {
    					matchedmoves.add(examinedmove);
    					moves.remove(j--);
    					continue;
    				}
    			}
    			
    			if (tactic.equals("moveuponfoundations") && !sourcepile.isstock) {
    				//Prioritize moves that move upwards on foundations.
    				if (targetpile.isfoundation && targetpile.getTopCard() != null && targetpile.getTopCard().value == examinedmove.movedcard.value - 1) {
    					matchedmoves.add(examinedmove);
    					moves.remove(j--);
    					continue;
    				}
    			}
    			
    			if (tactic.equals("samedirection") && targetpile.isfoundation && !sourcepile.isstock && targetpile.stackpointer > 1) {
    				Card lastcard = targetpile.getTopCard();
    				Card secondlastcard = targetpile.contents[targetpile.stackpointer - 2];
    				int ordervalue = lastcard.value - secondlastcard.value;
    				
    				//Ordervalue will be 1 if going up, -1 if going down.
    				if (examinedmove.movedcard.value == lastcard.value + ordervalue) {
    					matchedmoves.add(examinedmove);
    					moves.remove(j--);
    					continue;
    				}
    				else if (examinedmove.movedcard.value == (lastcard.value + ordervalue) - 13 && targetpile.oneaway) {
    					matchedmoves.add(examinedmove);
    					moves.remove(j--);
    					continue;
    				}
    				else if (examinedmove.movedcard.value == (lastcard.value + ordervalue) + 13 && targetpile.oneaway) {
    					matchedmoves.add(examinedmove);
    					moves.remove(j--);
    					continue;
    				}
    			}
    			
    			if (tactic.equals("avoidleavingspaces") && !sourcepile.isstock) {
    				//Prioritize moves that do not leave the stackpointer at 0.
    				if (sourcepile.stackpointer > 1) {
    					matchedmoves.add(examinedmove);
    					moves.remove(j--);
    					continue;
    				}
    			}
    			
    			if (tactic.equals("dontstartnewfoundations") && !sourcepile.isstock && targetpile.isfoundation && targetpile.stackpointer > 0) {
    				//Prioritize foundation moves that move to an already filled pile.
    				matchedmoves.add(examinedmove);
					moves.remove(j--);
					continue;
    			}
    			
    			if (tactic.equals("unstock") && sourcepile.isstock) {
    				matchedmoves.add(examinedmove);
					moves.remove(j--);
					continue;
    			}
    		}
    		
    		//Add the matched moves to the list of prioritized moves.
    		while (matchedmoves.size() > 0) {
    			Move removed = (Move)matchedmoves.removeLast();
        		prioritizedmoves.add(removed);
        	}
    		
    	}
    	
    	//Now put the remaining moves on the top of the stack.
    	while (moves.size() > 0) {
    		prioritizedmoves.add((Move)moves.remove(0));
    	}
    	
    	return (prioritizedmoves);
    }
}
